# Copyright (c) 2021-2022 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy import exceptions
from sushy.oem.dell.resources.system import constants as sys_cons
from sushy.oem.dell.resources.system import raid_service
from sushy.oem.dell.resources.system import system as oem_system


class SystemTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()

        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'system.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        self.oem_system = oem_system.DellSystemExtension(
            self.conn, '/redfish/v1/Systems/System.Embedded.1')

        mock_perc_raid = mock.Mock(
            identity='Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1')
        type(mock_perc_raid).volumes = mock.PropertyMock(
            side_effect=exceptions.MissingAttributeError)
        mock_perc_nonraid = mock.Mock(
            identity='Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1',
            volumes=[mock.Mock(volume_type='rawdevice', raid_type=None)])

        mock_boss_controller = mock.MagicMock(raid_types=['RAID1'])
        mock_boss_controller.name = 'BOSS-S1'
        mock_boss = mock.Mock(storage_controllers=[mock_boss_controller],
                              drives=mock.Mock())
        mock_perc_controller = mock.MagicMock(raid_types=['RAID1'])
        mock_perc_controller.name = 'PERC'
        mock_perc = mock.Mock(storage_controllers=[mock_perc_controller],
                              drives=[mock_perc_raid, mock_perc_nonraid])

        mock_system = mock.Mock()
        mock_storage_nocontroller = mock.Mock(storage_controllers=[])
        mock_storage_nonraid = mock.Mock(
            storage_controllers=[mock.Mock(raid_types=[])])
        mock_storage_boss = mock_boss
        mock_storage_raid = mock_perc
        mock_system.storage.get_members.return_value = [
            mock_storage_nocontroller, mock_storage_nonraid, mock_storage_boss,
            mock_storage_raid]
        self.oem_system._parent_resource = mock_system

    def test_raid_service(self):
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'raid_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        result = self.oem_system.raid_service
        self.assertEqual(
            '/redfish/v1/Systems/System.Embedded.1/Oem/Dell/DellRaidService',
            result.path)
        self.assertIsInstance(result,
                              raid_service.DellRaidService)

    def test_change_physical_disk_state_raid(self):
        mock_taskmon = mock.Mock()
        mock_raid = mock.Mock()
        mock_raid.return_value = mock_taskmon
        mock_nonraid = mock.Mock()
        self.oem_system.raid_service.convert_to_raid = mock_raid
        self.oem_system.raid_service.convert_to_nonraid = mock_nonraid

        task_mons = self.oem_system.change_physical_disk_state(
            sys_cons.PhysicalDiskStateMode.RAID)

        self.assertEqual([mock_taskmon], task_mons)
        mock_raid.assert_called_once_with(
            ['Disk.Bay.1:Enclosure.Internal.0-1:RAID.Integrated.1-1'])
        mock_nonraid.assert_not_called()

    def test_change_physical_disk_state_nonraid(self):
        mock_taskmon = mock.Mock()
        mock_raid = mock.Mock()
        mock_nonraid = mock.Mock()
        mock_nonraid.return_value = mock_taskmon
        self.oem_system.raid_service.convert_to_raid = mock_raid
        self.oem_system.raid_service.convert_to_nonraid = mock_nonraid

        task_mons = self.oem_system.change_physical_disk_state(
            sys_cons.PhysicalDiskStateMode.NONRAID)

        self.assertEqual([mock_taskmon], task_mons)
        mock_raid.assert_not_called()
        mock_nonraid.assert_called_once_with(
            ['Disk.Bay.0:Enclosure.Internal.0-1:RAID.Integrated.1-1'])

    def test_clear_foreign_config(self):
        mock_taskmon = mock.Mock()
        mock_clear_foreign_config = mock.Mock()
        mock_clear_foreign_config.side_effect = [mock_taskmon]
        self.oem_system.raid_service.clear_foreign_config =\
            mock_clear_foreign_config

        task_mons = self.oem_system.clear_foreign_config()

        self.assertEqual([mock_taskmon], task_mons)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

import sushy
from sushy import exceptions
from sushy.resources.certificateservice import certificate
from sushy.resources.certificateservice import certificateservice
from sushy.tests.unit import base


class CertificateServiceTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples'
                  '/certificateservice.json') as f:
            self.json_service = json.load(f)

        with open('sushy/tests/unit/json_samples'
                  '/certificate_locations.json') as f:
            self.json_locations = json.load(f)

        self.conn.get.return_value.json.side_effect = [
            self.json_service,
            self.json_locations,
        ]

        self.certificateservice = certificateservice.CertificateService(
            self.conn, '/redfish/v1/CertificateService',
            redfish_version='1.0.4')

    def test__parse_attributes(self):
        self.certificateservice._parse_attributes(self.json_service)

        self.assertEqual(self.certificateservice.identity,
                         'CertificateService')
        self.assertEqual(self.certificateservice.name, 'Certificate Service')
        self.assertEqual(self.certificateservice._certificate_locations_path,
                         '/redfish/v1/CertificateService/CertificateLocations')

    @mock.patch.object(certificate.Certificate, '__init__', autospec=True,
                       return_value=None)
    def test_ceritificate_locations(self, mock_cert):
        locs = self.certificateservice.certificate_locations
        self.assertEqual([
            "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1"
        ], locs.members_identities)
        cert = locs.get_member(
            "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1")
        self.assertIsInstance(cert, certificate.Certificate)
        mock_cert.assert_called_once_with(
            cert,
            self.conn,
            "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1",
            redfish_version='1.0.4',
            registries=self.certificateservice.registries,
            root=self.certificateservice.root)

    def test_replace_certificate(self):
        loc = "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1"
        self.certificateservice.replace_certificate(
            loc, 'abcd', sushy.CertificateType.PEM)
        self.conn.post.assert_called_once_with(
            "/redfish/v1/CertificateService/Actions"
            "/CertificateService.ReplaceCertificate",
            data={'CertificateUri': loc,
                  'CertificateString': 'abcd',
                  'CertificateType': 'PEM'})

    def test_replace_certificate_wrong_type(self):
        loc = "/redfish/v1/Managers/BMC/NetworkProtocol/HTTPS/Certificates/1"
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.certificateservice.replace_certificate,
                          loc, 'abcd', 'Scrum Master Certificate')



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json
import time
from unittest import mock

import requests

from sushy import auth as sushy_auth
from sushy import connector
from sushy import exceptions
from sushy.tests.unit import base


class ConnectorMethodsTestCase(base.TestCase):

    @mock.patch.object(sushy_auth, 'SessionOrBasicAuth', autospec=True)
    def setUp(self, mock_auth):
        mock_auth.get_session_key.return_value = None
        super().setUp()
        self.conn = connector.Connector(
            'http://foo.bar:1234', verify=True)
        self.conn._auth = mock_auth
        self.data = {'fake': 'data'}
        self.headers = {'X-Fake': 'header'}

    def test_init_with_credentials(self):
        conn = connector.Connector('http://foo.bar:1234',
                                   username='admin',
                                   password='password')
        self.assertEqual(conn._session.auth, ('admin', 'password'))

    def test_init_with_callback(self):
        def response_callback(response):
            return

        conn = connector.Connector('http://foo.bar:1234',
                                   username='admin',
                                   password='password',
                                   response_callback=response_callback)
        self.assertIs(conn._response_callback, response_callback)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_get(self, mock__op):
        self.conn.get(path='fake/path', data=self.data.copy(),
                      headers=self.headers.copy())
        mock__op.assert_called_once_with(mock.ANY, 'GET', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=False, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_get_blocking(self, mock__op):
        self.conn.get(path='fake/path', data=self.data.copy(),
                      headers=self.headers.copy(), blocking=True)
        mock__op.assert_called_once_with(mock.ANY, 'GET', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=True, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_post(self, mock__op):
        self.conn.post(path='fake/path', data=self.data.copy(),
                       headers=self.headers.copy())
        mock__op.assert_called_once_with(mock.ANY, 'POST', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=False, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_post_blocking(self, mock__op):
        self.conn.post(path='fake/path', data=self.data.copy(),
                       headers=self.headers.copy(), blocking=True, timeout=120)
        mock__op.assert_called_once_with(mock.ANY, 'POST', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=True, timeout=120)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_patch(self, mock__op):
        self.conn.patch(path='fake/path', data=self.data.copy(),
                        headers=self.headers.copy())
        mock__op.assert_called_once_with(mock.ANY, 'PATCH', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=False, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_patch_blocking(self, mock__op):
        self.conn.patch(path='fake/path', data=self.data.copy(), timeout=120,
                        headers=self.headers.copy(), blocking=True)
        mock__op.assert_called_once_with(mock.ANY, 'PATCH', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=True, timeout=120)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_put(self, mock__op):
        self.conn.put(path='fake/path', data=self.data.copy(),
                      headers=self.headers.copy())
        mock__op.assert_called_once_with(mock.ANY, 'PUT', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=False, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_put_blocking(self, mock__op):
        self.conn.put(path='fake/path', data=self.data.copy(), timeout=120,
                      headers=self.headers.copy(), blocking=True)
        mock__op.assert_called_once_with(mock.ANY, 'PUT', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=True, timeout=120)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_delete(self, mock__op):
        self.conn.delete(path='fake/path', data=self.data.copy(),
                         headers=self.headers.copy())
        mock__op.assert_called_once_with(mock.ANY, 'DELETE', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=False, timeout=None)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_delete_blocking(self, mock__op):
        self.conn.delete(path='fake/path', data=self.data.copy(), timeout=120,
                         headers=self.headers.copy(), blocking=True)
        mock__op.assert_called_once_with(mock.ANY, 'DELETE', 'fake/path',
                                         data=self.data, headers=self.headers,
                                         blocking=True, timeout=120)

    def test_set_auth(self):
        mock_auth = mock.MagicMock()
        self.conn.set_auth(mock_auth)
        self.assertEqual(mock_auth, self.conn._auth)

    def test_set_http_basic_auth(self):
        self.conn.set_http_basic_auth('foo', 'secret')
        self.assertEqual(('foo', 'secret'), self.conn._session.auth)

    def test_set_http_session_auth(self):
        self.conn.set_http_session_auth('hash-token')
        self.assertIn('X-Auth-Token', self.conn._session.headers)
        self.assertEqual(
            'hash-token', self.conn._session.headers['X-Auth-Token'])

    def test_close(self):
        session = mock.Mock(spec=requests.Session)
        self.conn._session = session
        self.conn.close()
        session.close.assert_called_once_with()


class ConnectorOpTestCase(base.TestCase):

    @mock.patch.object(sushy_auth, 'SessionOrBasicAuth', autospec=True)
    def setUp(self, mock_auth):
        mock_auth.get_session_key.return_value = None
        mock_auth._session_key = None
        self.auth = mock_auth
        super().setUp()
        self.conn = connector.Connector(
            'http://foo.bar:1234', verify=True,
            server_side_retries=10, server_side_retries_delay=3)
        self.conn._auth = mock_auth
        self.data = {'fake': 'data'}
        self.headers = {'Accept-Encoding': 'identity', 'OData-Version': '4.0'}
        self.session = mock.Mock(spec=requests.Session)
        self.conn._session = self.session
        self.request = self.session.request
        self.request.return_value.status_code = http_client.OK

    def test_ok_get(self):
        self.conn._op('GET', path='fake/path')
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234/fake/path',
            headers=self.headers, json=None, verify=True, timeout=60)

    def test_ok_get_with_headers(self):
        self.conn._op('GET', path='fake/path', headers={'answer': '42'})
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234/fake/path',
            headers=dict(self.headers, answer='42'),
            json=None, verify=True, timeout=60)

    def test_ok_get_default_timeout(self):
        self.conn._default_request_timeout = 42
        self.conn._op('GET', path='fake/path')
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234/fake/path',
            headers=self.headers, json=None, verify=True, timeout=42)

    def test_ok_get_connect_timeout_tuple(self):
        self.conn._default_request_timeout = 60
        self.conn._connect_timeout = 10
        self.conn._op('GET', path='fake/path')
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234/fake/path',
            headers=self.headers, json=None, verify=True, timeout=(10, 60))

    def test_response_callback(self):
        mock_response_callback = mock.MagicMock()
        self.conn._response_callback = mock_response_callback

        self.conn._op('GET', path='fake/path', headers=self.headers)
        self.assertEqual(1, mock_response_callback.call_count)

    def test_ok_get_url_redirect_false(self):
        self.conn._op('GET', path='fake/path', allow_redirects=False)
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234/fake/path',
            headers=self.headers, json=None, allow_redirects=False,
            verify=True, timeout=60)

    def test_ok_post(self):
        self.conn._op('POST', path='fake/path', data=self.data.copy())
        self.request.assert_called_once_with(
            'POST', 'http://foo.bar:1234/fake/path',
            json=self.data,
            headers=dict(self.headers, **{'Content-Type': 'application/json'}),
            verify=True, timeout=60)

    def test_ok_post_with_headers(self):
        self.conn._op('POST', path='fake/path', data=self.data.copy(),
                      headers={'answer': 42})
        self.request.assert_called_once_with(
            'POST', 'http://foo.bar:1234/fake/path',
            json=self.data,
            headers=dict(self.headers, **{'Content-Type': 'application/json',
                                          'answer': 42}),
            verify=True, timeout=60)

    def test_ok_put(self):
        self.conn._op('PUT', path='fake/path', data=self.data.copy())
        self.request.assert_called_once_with(
            'PUT', 'http://foo.bar:1234/fake/path',
            headers=dict(self.headers, **{'Content-Type': 'application/json'}),
            json=self.data, verify=True, timeout=60)

    def test_ok_delete(self):
        self.conn._op('DELETE', path='fake/path')
        self.request.assert_called_once_with(
            'DELETE', 'http://foo.bar:1234/fake/path',
            headers=self.headers, json=None, verify=True, timeout=60)

    def test_ok_post_with_session(self):
        self.conn._session.headers = {}
        self.conn._session.headers['X-Auth-Token'] = 'asdf1234'
        expected_headers = self.headers.copy()
        expected_headers['OData-Version'] = '4.0'
        expected_headers['Content-Type'] = 'application/json'
        self.conn._op('POST', path='fake/path', headers=self.headers,
                      data=self.data)
        self.request.assert_called_once_with(
            'POST', 'http://foo.bar:1234/fake/path',
            json=self.data, headers=expected_headers, verify=True, timeout=60)
        self.assertEqual(self.conn._session.headers,
                         {'X-Auth-Token': 'asdf1234'})

    def test_odata_version_header_redfish(self):
        path = '/redfish/v1/path'
        headers = dict(self.headers)
        expected_headers = dict(self.headers)
        expected_headers['OData-Version'] = '4.0'
        self.request.reset_mock()
        self.conn._op('GET', path=path, headers=headers)
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234' + path,
            headers=expected_headers, json=None, verify=True, timeout=60)

    def test_odata_version_header_redfish_existing_header(self):
        path = '/redfish/v1/foo'
        headers = {'OData-Version': '3.0'}
        expected_headers = dict(self.headers, **headers)
        self.conn._op('GET', path=path, headers=headers)
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234' + path,
            headers=expected_headers, json=None, verify=True, timeout=60)

    def test_remove_header_accept_encoding(self):
        path = '/redfish/v1/foo'
        headers = {'Accept-Encoding': None}
        self.headers.pop('Accept-Encoding')
        self.conn._op('GET', path=path, headers=headers)
        self.request.assert_called_once_with(
            'GET', 'http://foo.bar:1234' + path,
            headers=self.headers, json=None, verify=True, timeout=60)

    def test_timed_out_session_unable_to_create_session(self):
        self.conn._auth.can_refresh_session.return_value = False
        self.session.auth = None
        self.conn._session = self.session
        self.request = self.session.request
        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = ValueError('no json')
        mock_response = mock.Mock()
        mock_response.json.side_effect = ValueError('no json')
        mock_response.status_code = http_client.FORBIDDEN
        self.conn._auth.authenticate.side_effect = \
            exceptions.AccessError('POST', 'fake/path', mock_response)
        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as ae:
            self.conn._op('POST', path='fake/path', data=self.data,
                          headers=self.headers)
        exc = ae.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)

    def test_timed_out_session_re_established(self):
        self.auth._session_key = 'asdf1234'
        self.auth.get_session_key.return_value = 'asdf1234'
        self.conn._auth = self.auth
        self.session = mock.Mock(spec=requests.Session)
        self.session.auth = None
        self.conn._session = self.session
        self.request = self.session.request
        first_response = mock.MagicMock()
        first_response.status_code = http_client.FORBIDDEN
        second_response = mock.MagicMock()
        second_response.status_code = http_client.OK
        second_response.json = {'Test': 'Testing'}
        self.request.side_effect = [first_response, second_response]
        response = self.conn._op('POST', path='fake/path', data=self.data,
                                 headers=self.headers)
        self.auth.refresh_session.assert_called_with()
        self.auth.can_refresh_session.assert_called_with()
        self.assertEqual(response.json, second_response.json)

    def test_timed_out_session_timed_out_refresh(self):
        self.auth._session_key = 'asdf1234'
        self.auth.get_session_key.return_value = 'asdf1234'
        self.conn._auth = self.auth
        self.session = mock.Mock(spec=requests.Session)
        self.session.auth = None
        self.conn._session = self.session
        self.request = self.session.request
        first_response = mock.MagicMock()
        first_response.status_code = http_client.FORBIDDEN
        second_response = first_response
        self.auth.refresh_session.side_effect = \
            exceptions.ConnectionError('meow')
        third_response = mock.MagicMock()
        third_response.status_code = http_client.OK
        third_response.json = {'Test': 'Testing'}
        self.auth.can_refresh_session.return_value = True

        self.request.side_effect = [first_response, second_response,
                                    third_response]
        self.assertRaises(exceptions.ConnectionError, self.conn._op, 'POST',
                          path='fake/path', data=self.data,
                          headers=self.headers)
        self.auth.refresh_session.assert_called_with()
        self.auth.refresh_session.reset_mock()
        # Normally, this would be reset by refresh_session, but given
        # the heavy mocking, we need to do it for this test.
        self.auth._session_key = None
        self.auth.get_session_key.return_value = None
        self.auth.can_refresh_session.return_value = False

        response = self.conn._op('POST', path='fake/path', data=self.data,
                                 headers=self.headers)
        self.auth.refresh_session.assert_not_called()
        self.auth.authenticate.assert_called_once()
        self.assertEqual(response.json, third_response.json)

    def test_timed_out_session_fail_after_reestablish(self):
        self.auth._session_key = 'asdf1234'
        self.auth.get_session_key.return_value = 'asdf1234'
        self.conn._auth = self.auth
        self.session = mock.Mock(spec=requests.Session)
        self.session.auth = None
        self.conn._session = self.session
        self.request = self.session.request
        first_response = mock.MagicMock()
        first_response.status_code = http_client.FORBIDDEN
        second_response = mock.MagicMock()
        second_response.status_code = http_client.BAD_REQUEST
        self.request.side_effect = [first_response, second_response]
        self.assertRaises(exceptions.BadRequestError,
                          self.conn._op,
                          'POST', path='fake/path', data=self.data,
                          headers=self.headers)
        self.auth.refresh_session.assert_called_with()
        self.auth.can_refresh_session.assert_called_with()

    def test_timed_out_session_fail_after_reestablish_no_recursion(self):
        self.auth._session_key = 'asdf1234'
        self.auth.get_session_key.return_value = 'asdf1234'
        self.conn._auth = self.auth
        self.session = mock.Mock(spec=requests.Session)
        self.session.auth = None
        self.conn._session = self.session
        self.request = self.session.request
        first_response = mock.MagicMock()
        first_response.status_code = http_client.FORBIDDEN
        second_response = mock.MagicMock()
        second_response.status_code = http_client.FORBIDDEN
        self.request.side_effect = [first_response, second_response]
        self.assertRaises(exceptions.AccessError,
                          self.conn._op,
                          'POST', path='fake/path', data=self.data,
                          headers=self.headers)
        self.auth.refresh_session.assert_called_with()
        self.auth.can_refresh_session.assert_called_with()

    @mock.patch.object(time, 'sleep', autospec=True)
    def test_connection_error(self, mock_sleep):
        self.request.side_effect = requests.exceptions.ConnectionError
        self.assertRaises(exceptions.ConnectionError, self.conn._op, 'GET')

    def test_unknown_http_error(self):
        self.request.return_value.status_code = http_client.CONFLICT
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.HTTPError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.CONFLICT, exc.status_code)
        self.assertIsNone(exc.body)
        self.assertIsNone(exc.detail)

    def test_known_http_error(self):
        self.request.return_value.status_code = http_client.BAD_REQUEST
        with open('sushy/tests/unit/json_samples/error.json') as f:
            self.request.return_value.json.return_value = json.load(f)

        with self.assertRaisesRegex(exceptions.BadRequestError,
                                    'body submitted was malformed JSON') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.BAD_REQUEST, exc.status_code)
        self.assertIsNotNone(exc.body)
        self.assertIn('body submitted was malformed JSON', exc.detail)
        self.assertEqual(len(exc.extended_info), 3, exc.extended_info)

    def test_known_http_error_nonlist_ext_info(self):
        self.request.return_value.status_code =\
            http_client.UNSUPPORTED_MEDIA_TYPE
        with open('sushy/tests/unit/json_samples/'
                  'error_single_ext_info.json') as f:
            self.request.return_value.json.return_value = json.load(f)

        with self.assertRaisesRegex(exceptions.HTTPError,
                                    'See Resolution for information') as cm:
            self.conn._op('POST', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.UNSUPPORTED_MEDIA_TYPE, exc.status_code)
        self.assertIsNotNone(exc.body)
        self.assertIn('See Resolution for information', exc.detail)
        self.assertIn('unsupported media type',
                      exc.extended_info['Resolution'])

    @mock.patch('time.sleep', autospec=True)
    def test_not_found_error(self, mock_sleep):
        self.request.return_value.status_code = http_client.NOT_FOUND
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.ResourceNotFoundError,
                                    'Resource http://foo.bar not found') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.NOT_FOUND, exc.status_code)
        self.assertFalse(mock_sleep.called)
        self.assertEqual(1, self.request.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test_server_error(self, mock_sleep):
        self.request.return_value.status_code = (
            http_client.INTERNAL_SERVER_ERROR)
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.ServerSideError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.INTERNAL_SERVER_ERROR, exc.status_code)
        self.assertEqual(10, mock_sleep.call_count)
        self.assertEqual(11, self.request.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test_op_retry_on_server_500_sys518(self, mock_sleep):
        response_info = {"error": {"@Message.ExtendedInfo": [
            {'MessageId': 'IDRAC.2.7.SYS518'}]}}
        mock_error = mock.Mock()
        mock_error.status_code = 500
        mock_error.json.return_value = response_info
        self.request.return_value.status_code = (
            http_client.INTERNAL_SERVER_ERROR)
        self.request.return_value.json.side_effect =\
            exceptions.ServerSideError(
                method='DELETE', url='http://foo.bar', response=mock_error)

        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
                          'http://foo.bar')
        self.assertEqual(10, mock_sleep.call_count)
        self.assertEqual(11, self.request.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test_op_retry_on_server_500_other_than_sys518(self, mock_sleep):
        response_info = {"error": {"@Message.ExtendedInfo": [
            {'MessageId': 'IDRAC.2.7.SYS999'}]}}
        mock_error = mock.Mock()
        mock_error.status_code = 500
        mock_error.json.return_value = response_info
        self.request.return_value.status_code = (
            http_client.INTERNAL_SERVER_ERROR)
        self.request.return_value.json.side_effect =\
            exceptions.ServerSideError(
                method='DELETE', url='http://foo.bar', response=mock_error)

        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
                          'http://foo.bar')
        self.assertEqual(0, mock_sleep.call_count)
        self.assertEqual(1, self.request.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test_op_retry_on_server_400_ilo_not_ready(self, mock_sleep):
        response_info = {"error": {"@Message.ExtendedInfo": [
            {'MessageId': 'iLO.2.15.InvalidOperationForSystemState'}]}}
        mock_error = mock.Mock()
        mock_error.status_code = 400
        mock_error.json.return_value = response_info
        self.request.return_value.status_code = (
            http_client.INTERNAL_SERVER_ERROR)
        self.request.return_value.json.side_effect =\
            exceptions.ServerSideError(
                method='DELETE', url='http://foo.bar', response=mock_error)

        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
                          'http://foo.bar')
        self.assertEqual(10, mock_sleep.call_count)
        self.assertEqual(11, self.request.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test_op_retry_on_server_400_ilo_not_ready_other_error(self,
                                                              mock_sleep):
        response_info = {"error": {"@Message.ExtendedInfo": [
            {'MessageId': 'iLO.Invalid'}]}}
        mock_error = mock.Mock()
        mock_error.status_code = 400
        mock_error.json.return_value = response_info
        self.request.return_value.status_code = (
            http_client.INTERNAL_SERVER_ERROR)
        self.request.return_value.json.side_effect =\
            exceptions.ServerSideError(
                method='DELETE', url='http://foo.bar', response=mock_error)

        self.assertRaises(exceptions.ServerSideError, self.conn._op, 'DELETE',
                          'http://foo.bar')
        self.assertEqual(0, mock_sleep.call_count)
        self.assertEqual(1, self.request.call_count)

    def test_op_retry_without_identity(self):
        self.request.side_effect = [
            mock.Mock(status_code=http_client.NOT_ACCEPTABLE),
            mock.Mock(status_code=http_client.OK),
        ]
        self.conn._op('GET', 'http://foo.bar')

        self.assertEqual(2, self.request.call_count)
        headers_no_accept = self.headers.copy()
        headers_no_accept.pop('Accept-Encoding')
        self.request.assert_has_calls([
            mock.call('GET', 'http://foo.bar', headers=self.headers,
                      json=None, verify=True, timeout=60),
            mock.call('GET', 'http://foo.bar', headers=headers_no_accept,
                      json=None, verify=True, timeout=60),
        ])

    def test_op_retry_without_identity_fails(self):
        self.request.return_value.status_code = http_client.NOT_ACCEPTABLE
        self.assertRaises(exceptions.NotAcceptableError, self.conn._op,
                          'GET', 'http://foo.bar')
        self.assertEqual(2, self.request.call_count)

    def test_access_error(self):
        self.conn._auth = None

        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)

    def test_access_error_on_session_post_does_not_retry(self):
        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = ValueError('no json')

        self.conn._sessions_uri = '/redfish/v1/SessionService/Sessions'

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('POST', 'http://foo.bar/redfish/v1/Session'
                                  'Service/Sessions',
                          data={'foo': 'bar'})
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)
        self.request.assert_called_once()

    def test_access_error_triggers_auth_attempt(self):
        self.conn._auth.can_refresh_session.return_value = False
        value_error = ValueError('no json')
        mock_response = mock.Mock()
        mock_response.json.side_effect = value_error
        mock_response.status_code = http_client.FORBIDDEN
        self.session.auth = None
        self.conn._session = self.session
        # This doesn't test/wire all way back through auth -> main
        # and ultimately we want to make sure we get an error all
        # the way back.
        self.conn._auth.authenticate.side_effect = \
            exceptions.AccessError('GET', '/', mock_response)
        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = value_error

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)
        self.conn._auth.authenticate.assert_called_once()

    def test_access_error_without_auth(self):
        self.conn._auth = None

        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://foo.bar')
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)

    @mock.patch.object(connector.LOG, 'debug', autospec=True)
    def test_access_error_service_session_reauth(self, mock_log):
        self.conn._auth.can_refresh_session.return_value = False

        self.request.return_value.status_code = http_client.FORBIDDEN
        mock_response = mock.Mock()
        mock_response.json.side_effect = ValueError('no json')
        mock_response.status_code = http_client.FORBIDDEN
        self.session.auth = None
        self.conn._session = self.session
        self.conn._auth.authenticate.side_effect = \
            exceptions.AccessError('POST', 'fake/path', mock_response)
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://redfish/v1/SessionService')
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)
        self.conn._auth.authenticate.assert_called_once()

    @mock.patch.object(connector.LOG, 'debug', autospec=True)
    def test_access_error_service_session_no_auth(self, mock_log):
        self.conn._auth = None

        self.request.return_value.status_code = http_client.FORBIDDEN
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://redfish/v1/SessionService')
        exc = cm.exception
        mock_log.assert_called_with(
            'HTTP GET of SessionService failed %s, '
            'this is expected prior to authentication', 'HTTP GET '
            'http://redfish/v1/SessionService returned code '
            f'{http_client.FORBIDDEN!s}. unknown error Extended '
            'information: None')
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)

    def test_blocking_no_location_header(self):
        self.request.return_value.status_code = http_client.ACCEPTED
        self.request.return_value.headers = {'retry-after': 5}
        with self.assertRaisesRegex(exceptions.ConnectionError,
                                    'status 202, but no Location header'):
            self.conn._op('POST', 'http://foo.bar', blocking=True)

    @mock.patch('sushy.connector.time.sleep', autospec=True)
    def test_blocking_task_fails(self, mock_sleep):
        response1 = mock.MagicMock(spec=requests.Response)
        response1.status_code = http_client.ACCEPTED
        response1.headers = {
            'Retry-After': 5,
            'Location': '/redfish/v1/taskmon/1',
            'Content-Length': 10
        }
        response1.json.return_value = {'Id': 3, 'Name': 'Test'}
        response2 = mock.MagicMock(spec=requests.Response)
        response2.status_code = http_client.BAD_REQUEST
        message = 'Unable to create Volume with given parameters'
        response2.json.return_value = {
            'error': {
                'message': message
            }
        }
        self.request.side_effect = [response1, response1, response2]
        with self.assertRaisesRegex(exceptions.BadRequestError, message):
            self.conn._op('POST', 'http://foo.bar', blocking=True)

    @mock.patch('sushy.connector.time.sleep', autospec=True)
    def test_blocking_with_connect_timeout_tuple(self, mock_sleep):
        # Test that TaskMonitor.wait() receives the read timeout (second value)
        # when timeout is a tuple (connect, read). This verifies that the
        # tuple doesn't cause a TypeError in TaskMonitor.wait().
        self.conn._connect_timeout = 10
        self.conn._default_request_timeout = 60
        response1 = mock.MagicMock(spec=requests.Response)
        response1.status_code = http_client.ACCEPTED
        response1.headers = {
            'Retry-After': 5,
            'Location': '/redfish/v1/taskmon/1',
            'Content-Length': 10
        }
        response1.json.return_value = {'Id': 3, 'Name': 'Test'}
        response2 = mock.MagicMock(spec=requests.Response)
        response2.status_code = http_client.OK
        response2.json.return_value = {}
        self.request.side_effect = [response1, response2]
        # This should not raise - the tuple timeout should be handled correctly
        # by extracting the read timeout for TaskMonitor.wait()
        self.conn._op('POST', 'http://foo.bar', blocking=True)
        # Verify the first request was called with the tuple timeout
        first_call = self.request.call_args_list[0]
        self.assertEqual(first_call[1]['timeout'], (10, 60))

    @mock.patch.object(connector.LOG, 'warning', autospec=True)
    def test_access_error_basic_auth(self, mock_log):
        self.conn._auth.can_refresh_session.return_value = False
        self.conn._session.auth = ('foo', 'bar')
        self.request.return_value.status_code = 403
        mock_response = mock.Mock()
        mock_response.json.side_effect = ValueError('no json')
        mock_response.status_code = 403
        self.request.return_value.json.side_effect = ValueError('no json')

        with self.assertRaisesRegex(exceptions.AccessError,
                                    'unknown error') as cm:
            self.conn._op('GET', 'http://redfish/v1/SessionService')
        exc = cm.exception
        self.assertEqual(http_client.FORBIDDEN, exc.status_code)
        self.conn._auth.authenticate.assert_not_called()
        error_msg = (f'HTTP GET http://redfish/v1/SessionService '
                     f'returned code {exc.status_code}. unknown error '
                     f'Extended information: None')
        mock_log.assert_called_once_with(
            "We have encountered an AccessError when using 'basic' "
            "authentication. %(err)s",
            {'err': error_msg}
        )

    @mock.patch.object(time, 'sleep', autospec=True)
    def test__op_raises_connection_error(self, mock_sleep):

        exception_list = [
            requests.exceptions.ChunkedEncodingError,
            requests.exceptions.ContentDecodingError,
            requests.exceptions.ReadTimeout,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.SSLError,
            requests.exceptions.ConnectionError,
            requests.exceptions.RequestException,
        ]
        for exc in exception_list:
            self.request.side_effect = exc
            self.assertRaises(exceptions.ConnectionError,
                              self.conn._op,
                              'POST',
                              'http://foo.bar',
                              blocking=True)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_strong_etag_ok(self, mock__op):
        self.headers['If-Match'] = '"3d7b8a7360bf2941d"'
        response = mock.MagicMock(spec=requests.Response)
        response.status_code = http_client.OK
        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        mock__op.return_value = response
        self.conn._etag_handler(path='/redfish/v1/Systems/1',
                                headers=self.headers, data=data,
                                etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_once_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=dict(self.headers, **{'If-Match': '"3d7b8a7360bf2941d"'}),
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_weak_etag_ok(self, mock__op):
        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
        response = mock.MagicMock(spec=requests.Response)
        response.status_code = http_client.OK
        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        mock__op.return_value = response
        self.conn._etag_handler(path='/redfish/v1/Systems/1',
                                headers=self.headers, data=data,
                                etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_once_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=dict(self.headers,
                         **{'If-Match': 'W/"3d7b8a7360bf2941d"'}),
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_no_etag_ok(self, mock__op):
        response = mock.MagicMock(spec=requests.Response)
        response.status_code = http_client.OK
        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        mock__op.return_value = response
        self.conn._etag_handler(path='/redfish/v1/Systems/1',
                                headers=self.headers, data=data,
                                blocking=False, timeout=60)
        mock__op.assert_called_once_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=self.headers,
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_weak_etag_fallback_to_strong(self, mock__op):
        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
        target_uri = '/redfish/v1/Systems/1'
        response_412 = mock.MagicMock(spec=requests.Response)
        response_412.status_code = http_client.PRECONDITION_FAILED
        response_200 = mock.MagicMock(spec=requests.Response)
        response_200.status_code = http_client.OK
        mock__op.side_effect = [
            exceptions.HTTPError(method='PATCH', url=target_uri,
                                 response=response_412),
            response_200]

        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        self.conn._etag_handler(path=target_uri, headers=self.headers,
                                data=data, etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=dict(self.headers, **{'If-Match': '"3d7b8a7360bf2941d"'}),
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_weak_etag_fallback_to_no_etag(self, mock__op):
        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
        target_uri = '/redfish/v1/Systems/1'
        response_412 = mock.MagicMock(spec=requests.Response)
        response_412.status_code = http_client.PRECONDITION_FAILED
        response_200 = mock.MagicMock(spec=requests.Response)
        response_200.status_code = http_client.OK
        mock__op.side_effect = [
            exceptions.HTTPError(method='PATCH', url=target_uri,
                                 response=response_412),
            exceptions.HTTPError(method='PATCH', url=target_uri,
                                 response=response_412),
            response_200]

        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        self.conn._etag_handler(path=target_uri, headers=self.headers,
                                data=data, etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=self.headers,
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_strong_etag_fallback_to_no_etag(self, mock__op):
        self.headers['If-Match'] = '"3d7b8a7360bf2941d"'
        target_uri = '/redfish/v1/Systems/1'
        response_412 = mock.MagicMock(spec=requests.Response)
        response_412.status_code = http_client.PRECONDITION_FAILED
        response_200 = mock.MagicMock(spec=requests.Response)
        response_200.status_code = http_client.OK
        mock__op.side_effect = [
            exceptions.HTTPError(method='PATCH', url=target_uri,
                                 response=response_412),
            response_200]

        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        self.conn._etag_handler(path=target_uri, headers=self.headers,
                                data=data, etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=self.headers,
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_emptystring_etag(self, mock__op):
        self.headers_no_etag = self.headers.copy()
        target_uri = '/redfish/v1/Systems/1'
        response_200 = mock.MagicMock(spec=requests.Response)
        response_200.status_code = http_client.OK

        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        self.conn._etag_handler(path=target_uri, headers=self.headers,
                                data=data, etag='',
                                blocking=False, timeout=60)
        mock__op.assert_called_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=self.headers_no_etag,
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_emptystring_etag_from_headers(self, mock__op):
        self.headers_no_etag = self.headers.copy()
        self.headers['If-Match'] = ''
        target_uri = '/redfish/v1/Systems/1'
        response_200 = mock.MagicMock(spec=requests.Response)
        response_200.status_code = http_client.OK

        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        self.conn._etag_handler(path=target_uri, headers=self.headers,
                                data=data, etag=self.headers['If-Match'],
                                blocking=False, timeout=60)
        mock__op.assert_called_with(
            self.conn,
            "PATCH",
            '/redfish/v1/Systems/1',
            data={'Boot': {'BootSourceOverrideTarget': 'Cd',
                           'BootSourceOverrideEnabled': 'Once'}},
            headers=self.headers_no_etag,
            blocking=False,
            timeout=60)

    @mock.patch.object(connector.Connector, '_op', autospec=True)
    def test_etag_handler_fail_other_exception(self, mock__op):
        self.headers['If-Match'] = 'W/"3d7b8a7360bf2941d"'
        target_uri = '/redfish/v1/Systems/1'
        data = {'Boot': {'BootSourceOverrideTarget': 'Cd',
                         'BootSourceOverrideEnabled': 'Once'}}
        response = mock.MagicMock(spec=requests.Response)
        response.status_code = http_client.FORBIDDEN
        response.message = "boom"
        mock__op.return_value = response
        mock__op.side_effect = \
            exceptions.HTTPError(method='PATCH',
                                 url=target_uri,
                                 response=response)
        self.assertRaises(exceptions.HTTPError, self.conn._etag_handler,
                          'PATCH', target_uri, data,
                          self.headers,
                          blocking=False, timeout=60)

    @mock.patch.object(time, 'sleep', autospec=True)
    def test_retry_on_connection_error(self, mock_sleep):
        self.request.side_effect = [
            requests.exceptions.ConnectionError("temporary issue"),
            mock.Mock(status_code=200, headers={}, json=lambda: {})
        ]

        target_uri = '/redfish/v1/Systems/1'
        result = self.conn.get(target_uri)
        self.assertEqual(result.status_code, 200)
        self.assertEqual(self.request.call_count, 2)

    @mock.patch.object(time, 'sleep', autospec=True)
    def test_raises_after_max_retries(self, mock_sleep):
        target_uri = '/redfish/v1/Systems/1'
        self.request.side_effect = requests.exceptions.ConnectTimeout(
            "timeout")
        self.assertRaises(exceptions.ConnectionError, self.conn.get,
                          target_uri)
        self.assertEqual(self.request.call_count,
                         self.conn._server_side_retries)



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class JobState(enum.Enum):
    """Job state constants"""

    COMPLETED = "Completed"
    """A job is in completed state"""

    COMPLETED_ERRORS = "CompletedWithErrors"
    """A job is in completed state with errors"""

    DOWNLOADED = "Downloaded"
    """A job is in downloaded state"""

    DOWNLOADING = "Downloading"
    """A job is in downloading state"""

    FAILED = "Failed"
    """A job is in failed state"""

    NEW = "New"
    """A job is in newly created state"""

    PAUSED = "Paused"
    """A job is in paused state"""

    PENDING_ACTIVATION = "PendingActivation"
    """A job is in pending activation state"""

    READY_EXECUTION = "ReadyForExecution"
    """A job is in ready for execution state"""

    REBOOT_COMPLETED = "RebootCompleted"
    """A job is in reboot completed state"""

    REBOOT_FAILED = "RebootFailed"
    """A job is in reboot failed state"""

    REBOOT_PENDING = "RebootPending"
    """A job is in pending state for reboot"""

    RUNNING = "Running"
    """A job is in running state"""

    SCHEDULED = "Scheduled"
    """A job is in scheduled state"""

    SCHEDULING = "Scheduling"
    """A job is in scheduling state"""

    UNKNOWN = "Unknown"
    """A job is in unknown state"""

    WAITING = "Waiting"
    """A job is in waiting state"""


# Backward compatibility
JOB_STATE_COMPLETED = JobState.COMPLETED
JOB_STATE_COMPLETED_ERRORS = JobState.COMPLETED_ERRORS
JOB_STATE_DOWNLOADED = JobState.DOWNLOADED
JOB_STATE_DOWNLOADING = JobState.DOWNLOADING
JOB_STATE_FAILED = JobState.FAILED
JOB_STATE_NEW = JobState.NEW
JOB_STATE_PAUSED = JobState.PAUSED
JOB_STATE_PENDING_ACTIVATION = JobState.PENDING_ACTIVATION
JOB_STATE_READY_EXECUTION = JobState.READY_EXECUTION
JOB_STATE_REBOOT_COMPLETED = JobState.REBOOT_COMPLETED
JOB_STATE_REBOOT_FAILED = JobState.REBOOT_FAILED
JOB_STATE_REBOOT_PENDING = JobState.REBOOT_PENDING
JOB_STATE_RUNNING = JobState.RUNNING
JOB_STATE_SCHEDULED = JobState.SCHEDULED
JOB_STATE_SCHEDULING = JobState.SCHEDULING
JOB_STATE_UNKNOWN = JobState.UNKNOWN
JOB_STATE_WAITING = JobState.WAITING


class JobType(enum.Enum):
    """Job type constants"""

    BIOS_CONF = "BIOSConfiguration"
    """A BIOS configuration job"""

    EXPORT_CONF = "ExportConfiguration"
    """A server configuration profile export job"""

    FC_CONF = "FCConfiguration"
    """A Fibre Channel configuration job"""

    FACTORY_CONF_EXPORT = "FactoryConfigurationExport"
    """A factory configuration export job"""

    FIRMWARE_ROLLBACK = "FirmwareRollback"
    """A firmware rollback job"""

    FIRMWARE_UPDATE = "FirmwareUpdate"
    """A firmware update job"""

    HW_INVENTORY_EXPORT = "HardwareInventoryExport"
    """A hardware inventory export job"""

    IMPORT_CONF = "ImportConfiguration"
    """A server configuration profile import job"""

    INBAND_BIOS_CONF = "InbandBIOSConfiguration"
    """An inband BIOS configuration job"""

    LC_CONF = "LCConfig"
    """A lifecycle controller attribute configuration job"""

    LC_EXPORT = "LCExport"
    """A lifecycle controller export job"""

    LC_LOG_EXPORT = "LCLogExport"
    """A lifecycle controller log export job"""

    LICENSE_EXPORT = "LicenseExport"
    """A license export job"""

    LICENSE_IMPORT = "LicenseImport"
    """A license import job"""

    MSG_REG_EXPORT = "MessageRegistryExport"
    """Export message registry report job"""

    NIC_CONF = "NICConfiguration"
    """A NIC configuration job"""

    OS_DEPLOY = "OSDeploy"
    """Operating System deploy job"""

    RAID_CONF = "RAIDConfiguration"
    """A RAID configuration job"""

    RT_NO_REBOOT_CONF = "RealTimeNoRebootConfiguration"
    """A real time configuration job without reboot"""

    REBOOT_FORCE = "RebootForce"
    """A reboot job with forced shutdown"""

    REBOOT_NO_FORCE = "RebootNoForce"
    """A graceful reboot job without forced shutdown"""

    REBOOT_POWER_CYCLE = "RebootPowerCycle"
    """A power cycle job"""

    REMOTE_DIAG = "RemoteDiagnostics"
    """A remote diagnostics job"""

    REPO_UPDATE = "RepositoryUpdate"
    """An update job from a repository"""

    SA_COL_EXP_HEALTH_DATA = "SACollectExportHealthData"
    """Support Assist collect and export health data job"""

    SA_COL_HEALTH_DATA = "SACollectHealthData"
    """Support Assist collect health data job"""

    SA_EXP_HEALTH_DATA = "SAExportHealthData"
    """Support Assist export health data job"""

    SA_ISM = "SAExposeISM"
    """Support Assist expose ISM installer package to host job"""

    SA_REG = "SARegistration"
    """Support Assist register iDRAC to Dell backend server job"""

    SEKM_REKEY = "SEKMRekey"
    """A Secure Enterprise Key Manager rekey job"""

    SEKM_STATUS_SET = "SEKMStatusSet"
    """A Secure Enterprise Key Manager status set job"""

    SHUTDOWN = "Shutdown"
    """A shutdown job"""

    SYS_ERASE = "SystemErase"
    """A selective system erase job"""

    SYS_INFO_CONF = "SystemInfoConfiguration"
    """A system info configuration job"""

    THERMAL_HIST_EXP = "ThermalHistoryExport"
    """A thermal history export job"""

    UNKNOWN = "Unknown"
    """An unknown job"""

    IDRAC_CONF = "iDRACConfiguration"
    """An iDRAC configuration job"""


# Backward compatibility
JOB_TYPE_BIOS_CONF = JobType.BIOS_CONF
JOB_TYPE_EXPORT_CONF = JobType.EXPORT_CONF
JOB_TYPE_FC_CONF = JobType.FC_CONF
JOB_TYPE_FACTORY_CONF_EXPORT = JobType.FACTORY_CONF_EXPORT
JOB_TYPE_FIRMWARE_ROLLBACK = JobType.FIRMWARE_ROLLBACK
JOB_TYPE_FIRMWARE_UPDATE = JobType.FIRMWARE_UPDATE
JOB_TYPE_HW_INVENTORY_EXPORT = JobType.HW_INVENTORY_EXPORT
JOB_TYPE_IMPORT_CONF = JobType.IMPORT_CONF
JOB_TYPE_INBAND_BIOS_CONF = JobType.INBAND_BIOS_CONF
JOB_TYPE_LC_CONF = JobType.LC_CONF
JOB_TYPE_LC_EXPORT = JobType.LC_EXPORT
JOB_TYPE_LC_LOG_EXPORT = JobType.LC_LOG_EXPORT
JOB_TYPE_LICENSE_EXPORT = JobType.LICENSE_EXPORT
JOB_TYPE_LICENSE_IMPORT = JobType.LICENSE_IMPORT
JOB_TYPE_MSG_REG_EXPORT = JobType.MSG_REG_EXPORT
JOB_TYPE_NIC_CONF = JobType.NIC_CONF
JOB_TYPE_OS_DEPLOY = JobType.OS_DEPLOY
JOB_TYPE_RAID_CONF = JobType.RAID_CONF
JOB_TYPE_RT_NO_REBOOT_CONF = JobType.RT_NO_REBOOT_CONF
JOB_TYPE_REBOOT_FORCE = JobType.REBOOT_FORCE
JOB_TYPE_REBOOT_NO_FORCE = JobType.REBOOT_NO_FORCE
JOB_TYPE_REBOOT_POWER_CYCLE = JobType.REBOOT_POWER_CYCLE
JOB_TYPE_REMOTE_DIAG = JobType.REMOTE_DIAG
JOB_TYPE_REPO_UPDATE = JobType.REPO_UPDATE
JOB_TYPE_SA_COL_EXP_HEALTH_DATA = JobType.SA_COL_EXP_HEALTH_DATA
JOB_TYPE_SA_COL_HEALTH_DATA = JobType.SA_COL_HEALTH_DATA
JOB_TYPE_SA_EXP_HEALTH_DATA = JobType.SA_EXP_HEALTH_DATA
JOB_TYPE_SA_ISM = JobType.SA_ISM
JOB_TYPE_SA_REG = JobType.SA_REG
JOB_TYPE_SEKM_REKEY = JobType.SEKM_REKEY
JOB_TYPE_SEKM_STATUS_SET = JobType.SEKM_STATUS_SET
JOB_TYPE_SHUTDOWN = JobType.SHUTDOWN
JOB_TYPE_SYS_ERASE = JobType.SYS_ERASE
JOB_TYPE_SYS_INFO_CONF = JobType.SYS_INFO_CONF
JOB_TYPE_THERMAL_HIST_EXP = JobType.THERMAL_HIST_EXP
JOB_TYPE_UNKNOWN = JobType.UNKNOWN
JOB_TYPE_IDRAC_CONF = JobType.IDRAC_CONF



#    Copyright (c) 2021 Anexia Internetdienstleistungs GmbH
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/NetworkDeviceFunction.v1_3_3.json

from sushy.resources import base
from sushy.resources import common
from sushy.resources.system.network import constants
from sushy.resources.system.network import port
from sushy import utils


class BootTargetsField(base.ListField):
    lun_id = base.Field("LUNID")
    """The logical unit number (LUN) ID from which to boot on the device"""

    priority = base.Field("BootPriority", adapter=utils.int_or_none)
    """The relative priority for this entry in the boot targets array."""

    wwpn = base.Field("WWPN")
    """The World Wide Port Name (WWPN) from which to boot."""


class VLANField(base.CompositeField):
    vlan_enabled = base.Field("VLANEnable", adapter=utils.bool_or_none)
    vlan_id = base.Field("VLANId", adapter=utils.int_or_none)


class ISCSIBootField(base.CompositeField):
    authentication_method = base.MappedField(
        'AuthenticationMethod', constants.NetworkAuthenticationMethod)
    """The configured capability of this network device function."""

    initiator_default_gateway = base.Field('InitiatorDefaultGateway')
    """The IPv6 or IPv4 iSCSI boot default gateway."""

    initiator_ip_address = base.Field('InitiatorIPAddress')
    """The IPv6 or IPv4 address of the iSCSI initiator."""

    initiator_netmask = base.Field('InitiatorNetmask')
    """The IPv6 or IPv4 netmask of the iSCSI boot initiator."""

    ip_address_type = base.MappedField(
        'IPAddressType', constants.IPAddressType)
    """The type of IP address being populated IP address fields."""

    primary_dns = base.Field('PrimaryDNS')
    """The IPv6 or IPv4 address of the primary DNS server."""

    primary_lun = base.Field('PrimaryLUN', adapter=utils.int_or_none)
    """The logical unit number (LUN) for the primary iSCSI boot target."""

    primary_target_ip_address = base.Field('PrimaryTargetIPAddress')
    """The IPv4 or IPv6 address for the primary iSCSI boot target."""

    primary_target_tcp_port = base.Field('PrimaryTargetTCPPort')
    """The TCP port for the primary iSCSI boot target."""

    primary_vlan_enabled = base.Field(
        'PrimaryVLANEnable',
        adapter=utils.bool_or_none
    )
    """An indication of whether the primary VLAN is enabled."""

    primary_vlan_id = base.Field("PrimaryVLANId", adapter=utils.int_or_none)
    """The 802.1q VLAN ID to use for iSCSI boot from the primary target."""

    secondary_dns = base.Field('SecondaryDNS')
    """The IPv6 or IPv4 address of the secondary DNS server."""

    secondary_lun = base.Field('SecondaryLUN', adapter=utils.int_or_none)
    """The logical unit number (LUN) for the secondary iSCSI boot target."""

    secondary_target_ip_address = base.Field('SecondaryTargetIPAddress')
    """The IPv4 or IPv6 address for the secondary iSCSI boot target."""

    secondary_target_tcp_port = base.Field('SecondaryTargetTCPPort')
    """The TCP port for the secondary iSCSI boot target."""

    secondary_vlan_enabled = base.Field(
        'SecondaryVLANEnable', adapter=utils.bool_or_none)
    """An indication of whether the secondary VLAN is enabled."""

    secondary_vlan_id = base.Field(
        "SecondaryVLANId",
        adapter=utils.int_or_none
    )
    """The 802.1q VLAN ID to use for iSCSI boot from the secondary target."""


class EthernetField(base.CompositeField):
    mac_address = base.Field("MACAddress")
    """The currently configured MAC address of the resource"""

    mtu_size = base.Field("MTUSize", adapter=utils.int_or_none)
    """The Maximum Transmission Unit (MTU) configured for this resource"""

    permanent_mac_address = base.Field("PermanentMACAddress")
    """The permanent MAC address assigned to this resource"""

    vlan = VLANField("VLAN")
    """The VLAN for this interface"""


class FibreChannelField(base.CompositeField):
    boot_targets = BootTargetsField("BootTargets")
    """An array of Fibre Channel boot targets configured for this resource."""


class NetworkDeviceFunction(base.ResourceBase):
    capabilities = base.MappedListField(
        'NetDevFuncCapabilities', constants.NetworkDeviceTechnology)
    """An array of capabilities for this network device function."""

    type = base.MappedField(
        'NetDevFuncType', constants.NetworkDeviceTechnology)
    """The configured capability of this network device function."""

    description = base.Field('Description')
    """The network device function description"""

    ethernet = EthernetField("Ethernet")
    """The Ethernet capabilities, status, and configuration values."""

    fibre_channel = FibreChannelField("FibreChannel")
    """The Fibre Channel capabilities, status, and configuration values."""

    identity = base.Field('Id', required=True)
    """Identifier for the network device function"""

    iscsi_boot = ISCSIBootField('iSCSIBoot')
    """
    The iSCSI boot capabilities, status, and configuration
    for a network device function.
    """

    max_virtual_functions = base.Field(
        'MaxVirtualFunctions',
        adapter=utils.int_or_none
    )
    """
    The number of virtual functions that are available
    for this network device function.
    """

    name = base.Field('Name', required=True)
    """The network device function name"""

    status = common.StatusField('Status')
    """The status of the resource"""

    @property
    @utils.cache_it
    def assignable_physical_ports(self):
        """An array of physical ports to which this resource may be assigned.

        Network ports to which this network device function may be assigned.

        :raises: MissingAttributeError if '@odata.id' field is missing.
        :returns: A list of `NetworkPort` instances
        """
        paths = utils.get_sub_resource_path_by(
            self, "AssignablePhysicalPorts", is_collection=True)

        return [port.NetworkPort(self._conn, path,
                                 redfish_version=self.redfish_version,
                                 registries=self.registries,
                                 root=self.root
                                 )
                for path in paths]


class NetworkDeviceFunctionCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return NetworkDeviceFunction



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class CertificateType(enum.Enum):
    PEM = 'PEM'
    """A Privacy Enhanced Mail (PEM)-encoded single certificate."""

    PEM_CHAIN = 'PEMchain'
    """A Privacy Enhanced Mail (PEM)-encoded certificate chain."""

    PKCS7 = 'PKCS7'
    """A Privacy Enhanced Mail (PEM)-encoded PKCS7 certificate."""


class CertificateUsageType(enum.Enum):
    USER = 'User'
    """This certificate is a user certificate like those associated with a
    manager account."""

    WEB = 'Web'
    """This certificate is a web or HTTPS certificate like those used for
    event destinations."""

    SSH = 'SSH'
    """This certificate is used for SSH."""

    DEVICE = 'Device'
    """This certificate is a device type certificate like those associated
    with SPDM and other standards."""

    PLATFORM = 'Platform'
    """This certificate is a platform type certificate like those associated
    with SPDM and other standards."""

    BIOS = 'BIOS'
    """This certificate is a BIOS certificate like those associated with
    UEFI."""


class KeyUsage(enum.Enum):
    DIGITAL_SIGNATURE = 'DigitalSignature'
    """Verifies digital signatures, other than signatures on certificates
    and CRLs."""

    NON_REPUDIATION = 'NonRepudiation'
    """Verifies digital signatures, other than signatures on certificates
    and CRLs, and provides a non-repudiation service that protects
    against the signing entity falsely denying some action."""

    KEY_ENCIPHERMENT = 'KeyEncipherment'
    """Enciphers private or secret keys."""

    DATA_ENCIPHERMENT = 'DataEncipherment'
    """Directly enciphers raw user data without an intermediate symmetric
    cipher."""

    KEY_AGREEMENT = 'KeyAgreement'
    """Key agreement."""

    KEY_CERT_SIGN = 'KeyCertSign'
    """Verifies signatures on public key certificates."""

    CRL_SIGNING = 'CRLSigning'
    """Verifies signatures on certificate revocation lists (CRLs)."""

    ENCIPHER_ONLY = 'EncipherOnly'
    """Enciphers data while performing a key agreement."""

    DECIPHER_ONLY = 'DecipherOnly'
    """Deciphers data while performing a key agreement."""

    SERVER_AUTHENTICATION = 'ServerAuthentication'
    """TLS WWW server authentication."""

    CLIENT_AUTHENTICATION = 'ClientAuthentication'
    """TLS WWW client authentication."""

    CODE_SIGNING = 'CodeSigning'
    """Signs downloadable executable code."""

    EMAIL_PROTECTION = 'EmailProtection'
    """Email protection."""

    TIMESTAMPING = 'Timestamping'
    """Binds the hash of an object to a time."""

    OCSP_SIGNING = 'OCSPSigning'
    """Signs OCSP responses."""



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# From http://redfish.dmtf.org/schemas/swordfish/v1/Volume.json

import enum


# FIXME(dtantsur): this is deprecated in favour of InitializeMethod
class VolumeInitializeType(enum.Enum):
    FAST = 'Fast'
    """The volume is prepared for use quickly, typically by erasing just the
    beginning and end of the space so that partitioning can be performed."""

    SLOW = 'Slow'
    """The volume is prepared for use slowly, typically by completely
    erasing the volume."""


# Backward compatibility
VOLUME_INIT_TYPE_FAST = VolumeInitializeType.FAST
VOLUME_INIT_TYPE_SLOW = VolumeInitializeType.SLOW


class VolumeType(enum.Enum):
    RAW_DEVICE = 'RawDevice'
    """The volume is a raw physical device without any RAID or other
    virtualization applied."""

    NON_REDUNDANT = 'NonRedundant'
    """The volume is a non-redundant storage device."""

    MIRRORED = 'Mirrored'
    """The volume is a mirrored device."""

    STRIPED_WITH_PARITY = 'StripedWithParity'
    """The volume is a device which uses parity to retain redundant
    information."""

    SPANNED_MIRRORS = 'SpannedMirrors'
    """The volume is a spanned set of mirrored devices."""

    SPANNED_STRIPES_WITH_PARITY = 'SpannedStripesWithParity'
    """The volume is a spanned set of devices which uses parity to retain
    redundant information."""


# Backward compatibility
VOLUME_TYPE_RAW_DEVICE = VolumeType.RAW_DEVICE
VOLUME_TYPE_NON_REDUNDANT = VolumeType.NON_REDUNDANT
VOLUME_TYPE_MIRRORED = VolumeType.MIRRORED
VOLUME_TYPE_STRIPED_WITH_PARITY = VolumeType.STRIPED_WITH_PARITY
VOLUME_TYPE_SPANNED_MIRRORS = VolumeType.SPANNED_MIRRORS
VOLUME_TYPE_SPANNED_STRIPES_WITH_PARITY = \
    VolumeType.SPANNED_STRIPES_WITH_PARITY


class RAIDType(enum.Enum):
    RAID0 = 'RAID0'
    """A placement policy where consecutive logical blocks of data are
    uniformly distributed across a set of independent storage devices
    without offering any form of redundancy."""

    RAID1 = 'RAID1'
    """A placement policy where each logical block of data is stored on more
    than one independent storage device."""

    RAID3 = 'RAID3'
    """A placement policy using parity-based protection where logical bytes
    of data are uniformly distributed across a set of independent storage
    devices and where the parity is stored on a dedicated independent
    storage device."""

    RAID4 = 'RAID4'
    """A placement policy using parity-based protection where logical blocks
    of data are uniformly distributed across a set of independent storage
    devices and where the parity is stored on a dedicated independent
    storage device."""

    RAID5 = 'RAID5'
    """A placement policy using parity-based protection for storing stripes
    of 'n' logical blocks of data and one logical block of parity across
    a set of 'n+1' independent storage devices where the parity and data
    blocks are interleaved across the storage devices."""

    RAID6 = 'RAID6'
    """A placement policy using parity-based protection for storing stripes
    of 'n' logical blocks of data and two logical blocks of independent
    parity across a set of 'n+2' independent storage devices where the
    parity and data blocks are interleaved across the storage devices."""

    RAID10 = 'RAID10'
    """A placement policy that creates a striped device (RAID 0) over a set
    of mirrored devices (RAID 1)."""

    RAID01 = 'RAID01'
    """A data placement policy that creates a mirrored device (RAID 1) over
    a set of striped devices (RAID 0)."""

    RAID6TP = 'RAID6TP'
    """A placement policy that uses parity-based protection for storing
    stripes of 'n' logical blocks of data and three logical blocks of
    independent parity across a set of 'n+3' independent storage devices
    where the parity and data blocks are interleaved across the storage
    devices."""

    RAID1E = 'RAID1E'
    """A placement policy that uses a form of mirroring implemented over a
    set of independent storage devices where logical blocks are
    duplicated on a pair of independent storage devices so that data is
    uniformly distributed across the storage devices."""

    RAID50 = 'RAID50'
    """A placement policy that uses a RAID 0 stripe set over two or more
    RAID 5 sets of independent storage devices."""

    RAID60 = 'RAID60'
    """A placement policy that uses a RAID 0 stripe set over two or more
    RAID 6 sets of independent storage devices."""

    RAID00 = 'RAID00'
    """A placement policy that creates a RAID 0 stripe set over two or more
    RAID 0 sets."""

    RAID10E = 'RAID10E'
    """A placement policy that uses a RAID 0 stripe set over two or more
    RAID 10 sets."""

    RAID1_TRIPLE = 'RAID1Triple'
    """A placement policy where each logical block of data is mirrored three
    times across a set of three independent storage devices."""

    RAID10_TRIPLE = 'RAID10Triple'
    """A placement policy that uses a striped device (RAID 0) over a set of
    triple mirrored devices (RAID 1Triple)."""

    NONE = 'None'
    """A placement policy with no redundancy at the device level."""


# Backward compatibility
RAID_TYPE_RAID0 = RAIDType.RAID0
RAID_TYPE_RAID1 = RAIDType.RAID1
RAID_TYPE_RAID3 = RAIDType.RAID3
RAID_TYPE_RAID4 = RAIDType.RAID4
RAID_TYPE_RAID5 = RAIDType.RAID5
RAID_TYPE_RAID6 = RAIDType.RAID6
RAID_TYPE_RAID10 = RAIDType.RAID10
RAID_TYPE_RAID01 = RAIDType.RAID01
RAID_TYPE_RAID6TP = RAIDType.RAID6TP
RAID_TYPE_RAID1E = RAIDType.RAID1E
RAID_TYPE_RAID50 = RAIDType.RAID50
RAID_TYPE_RAID60 = RAIDType.RAID60
RAID_TYPE_RAID00 = RAIDType.RAID00
RAID_TYPE_RAID10E = RAIDType.RAID10E
RAID_TYPE_RAID1Triple = RAIDType.RAID1_TRIPLE
RAID_TYPE_RAID10Triple = RAIDType.RAID10_TRIPLE
RAID_TYPE_NONE = RAIDType.NONE



# Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sushy
from sushy.oem.dell.resources.system.storage import constants as s_cons
from sushy.resources import base
from sushy.resources.oem import base as oem_base


class DellStorageController(base.CompositeField):

    controller_mode = base.MappedField('ControllerMode', s_cons.ControllerMode)
    """Mode of RAID controller"""


class DellStorageControllerExtension(oem_base.OEMResourceBase):
    dell_storage_controller = DellStorageController('DellStorageController')

    def convert_to_raid(self):
        """Converts to RAID mode if applicable

        If PERC 9 or PERC 10 controller is in non-RAID mode, then convert
        to RAID mode. No changes made for PERC 11 and above as they support
        only RAID mode, and BOSS controller as it does not have controller
        mode.
        :returns: TaskMonitor if controller mode changes applied and need to
            reboot, otherwise None
        """
        # Some controllers do not have controller_mode
        try:
            controller_mode = self.dell_storage_controller.controller_mode
        except AttributeError:
            return None

        # BOSS will have this empty, PERC will have something assigned
        if controller_mode and controller_mode != s_cons.ControllerMode.RAID:
            patch = {
                "Oem": {
                    "Dell": {
                        "DellStorageController": {
                            "ControllerMode":
                                s_cons.ControllerMode.RAID.value}}}}
            return self._parent_resource.update(
                patch, apply_time=sushy.ApplyTime.ON_RESET)


def get_extension(*args, **kwargs):
    return DellStorageControllerExtension



# Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

import sushy
from sushy.oem.dell.resources.system.storage import constants as ctrl_cons
from sushy.oem.dell.resources.system.storage import controller as oem_ctrl
from sushy.resources.system.storage import controller as sushy_constroller


class ControllerTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()

        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'storage_controller.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

            self.controller = sushy_constroller.StorageController(
                self.conn,
                '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1')
            self.oem_controller = oem_ctrl.DellStorageControllerExtension(
                self.conn,
                '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1')
            self.oem_controller = self.oem_controller.set_parent_resource(
                self.controller, 'Dell')

    def test_parse_attributes(self):
        self.assertEqual(
            ctrl_cons.ControllerMode.EHBA,
            self.oem_controller.dell_storage_controller.controller_mode)

    def test_convert_to_raid(self):
        mock_controller = mock.Mock()
        mock_task_monitor = mock.Mock()
        mock_controller.update.return_value = mock_task_monitor
        self.oem_controller._parent_resource = mock_controller

        res = self.oem_controller.convert_to_raid()

        self.assertEqual(mock_task_monitor, res)
        patch = {"Oem": {"Dell": {"DellStorageController": {
                 "ControllerMode": "RAID"}}}}
        mock_controller.update.assert_called_once_with(
            patch, apply_time=sushy.ApplyTime.ON_RESET)

    def test_convert_to_raid_already_raid(self):
        mock_controller = mock.Mock()
        self.oem_controller._parent_resource = mock_controller
        json = self.oem_controller.json
        json['Oem']['Dell']['DellStorageController']['ControllerMode'] = 'RAID'
        self.oem_controller.refresh()

        res = self.oem_controller.convert_to_raid()

        self.assertIsNone(res)
        mock_controller.update.assert_not_called()
